/*
 * Zaptel sysfs support (Internal Details) for Zapata Telephony interface
 *
 * Written by Maxime MENEAULT <maxmeneault@free.fr>
 *
 * Some functions copied from the linux kernel.
 *
 * Copyleft (>) 2007 Maxime MENEAULT <maxmeneault@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
 
/*
* PRIVATE HEADER for zaptel-sysfs.c 
*
* You can see it as an extension of zaptel-sysfs.c rather than a standard header file.
*
* That's why this header is not protected by ifndef (not meant to be included twice).
*
* This contains all the nasty things about linux kernel versions incompatibilities
* and all the handy macros to handle the linux device model in an easy manner for
* zaptel-sysfs
*
* Some functions (at least one) have been litteraly-copied from newer kernels
* so as to give some functionnality to older ones.
*
* You may add your nasty defines here so as to keep zaptel-sysfs.c clean.
*
* In particular having LINUX_VERSION_CODE checking in zaptel-sysfs.c
* is not allowed!
*/

/* some will be already included but it shouldn't hurt */
#include <linux/kernel.h> /* vsnprintf ... */
#include <linux/device.h> /* linux device model */
#include <linux/sysfs.h> /* don't ask */
#include <linux/stat.h> /* S_IRUGO ... */
#include <linux/kdev_t.h> /* MAJOR ... */
#include <linux/err.h> /* PTR_ERR ... */


/* === Linux Versionning === */

/* == handling of uevent/hotplug == */

/*
* @note this may need an add-on soon because as the time I speak
* the kernel is changing its add_uevent_var function prototype
*/


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
#	define zt_sysfs_add_uevent_var add_uevent_var
#	define CLASS_UEVENT_NAME uevent

#else

/* before 2.6.16 everything was hotplug */
#	define CLASS_UEVENT_NAME hotplug

#	if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
#		define zt_sysfs_add_uevent_var add_hotplug_env_var
#	else
/* before kernel 2.6.10 neither add_hotplug_env_var nor add_uevent_var did exist
* so here is a copy taken from lib/kobject_uevent.c (kernel 2.6.19)
*/
static int zt_sysfs_add_uevent_var(char **envp, int num_envp, int *cur_index,
		   char *buffer, int buffer_size, int *cur_len,
		   const char *format, ...)
{
	va_list args;

	/*
	 * We check against num_envp - 1 to make sure there is at
	 * least one slot left after we return, since kobject_uevent()
	 * needs to set the last slot to NULL.
	 */
	if (*cur_index >= num_envp - 1)
		return -ENOMEM;

	envp[*cur_index] = buffer + *cur_len;

	va_start(args, format);
	*cur_len += vsnprintf(envp[*cur_index],
			      max(buffer_size - *cur_len, 0),
			      format, args) + 1;
	va_end(args);

	if (*cur_len > buffer_size)
		return -ENOMEM;

	(*cur_index)++;
	return 0;
}
#	endif
#endif

/* class didn't have owner member before 2.6.13 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,13)
#	define CLASS_OWNER(...) __VA_ARGS__
#else
#	define CLASS_OWNER(...)
#endif

/* before 2.6.12 devt member of class_device didn't exist */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12)
#	define CLASS_DEVICE_HAVE_DEVT
#endif

/* before 2.6.15 release member of class_device didn't exist */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15)
#	define CLASS_DEVICE_RELEASE(...) __VA_ARGS__
#else
#	define CLASS_DEVICE_RELEASE(...)
#endif

/* before 2.6.15 parent member of class_device didn't exist 
* (however we tried to make our way without it)
*/



/* === Class Device handling === */

/*
*@brief a wrapper structure arround a class_device to remember some 
* interesting things
*/
struct zt_sysfs_class_device {
	int is_static; /*!< did we allocate it with kmalloc or is it static memory? */
	int type; 	  /*!< type of this class device one of ZT_CLASS_DEVICE_TYPE_* */
	struct class_device device; /*!< the class_device */
	  
#ifndef CLASS_DEVICE_HAVE_DEVT
	dev_t devt;	/*!< for older kernels */
#endif
};

#define to_zt_class_device(_pdev)\
	container_of(_pdev, struct zt_sysfs_class_device, device)


/* @brief Class device dev_t member handling 
* This gives us the compatibility even before simple_class was added.
*/
#ifdef CLASS_DEVICE_HAVE_DEVT

#	define CLASS_DEVICE_DEVT_HAVE(...) __VA_ARGS__

#	define CLASS_DEVICE_DEVT_HAVENT(...)

#	define zt_class_device_register(zcldev)\
		class_device_register(&((zcldev)->device))

#	define zt_class_device_unregister(zcldev)\
		class_device_unregister(&((zcldev)->device))
#else

#	define CLASS_DEVICE_DEVT_HAVE(...)

#	define CLASS_DEVICE_DEVT_HAVENT(...) __VA_ARGS__

static ssize_t zt_show_dev_t(struct class_device *class_dev, char *buf)
{
	struct zt_sysfs_class_device *zcldev = NULL;
	/* maybe unecessary checking but doesn't hurt */
	if (!class_dev)
		return 0;
	zcldev = to_zt_class_device(class_dev);
	return print_dev_t(buf, zcldev->devt);
}
/* Apparently "udev" needs a file called "dev" under each class_device representing
* a device that should go into /dev
*/
static struct class_device_attribute zt_class_device_attr_devt =
{
	.attr = {.name = "dev", .owner = THIS_MODULE, .mode = S_IRUGO},
	.show = zt_show_dev_t,
};

static int zt_class_device_register(struct zt_sysfs_class_device *zcldev)
{
	int retval = class_device_register(&zcldev->device);
    
	if (retval)
		return retval;
        
	if (MAJOR(zcldev->devt)) {
		retval = class_device_create_file(&zcldev->device, &zt_class_device_attr_devt);
		if (retval)
			goto error;
	}
    
	return 0;
error:
	class_device_unregister(&zcldev->device);
	return retval;
}

static void zt_class_device_unregister(struct zt_sysfs_class_device *zcldev)
{
	if (zcldev) {
		class_device_remove_file(&zcldev->device, &zt_class_device_attr_devt);
		class_device_unregister(&zcldev->device);
	}
}
#endif



/*
* @brief function to release the the only class of zaptel 
* (it is a no-op because the class is static memory)
*/
static void zt_sysfs_class_release_nop(struct class *cls)
{
}

/* @brief function to release class devices under the zaptel class */
static void zt_sysfs_class_device_release(struct class_device *class_dev)
{
	struct zt_sysfs_class_device *zcldev = NULL;
    
	if (!class_dev)
		return;
    
	zcldev = to_zt_class_device(class_dev);
    
	if (zcldev->is_static)
		return;
        
	kfree(zcldev);
}



/* === Statically allocated Class devices handling === */

/* @brief To terminate an array of class devices */
#define ZT_STATIC_CLASS_DEVICE_NULL \
	{.device = {.class = NULL}}

/* @brief Add a static class device to an array 
* 
* This class_device will be under the main zaptel class
*/     
#define ZT_DEFINE_STATIC_CLASS_DEVICE(_name, _dev)\
	{\
		.is_static = 1,\
		.type = ZT_CLASS_DEVICE_TYPE_DEFAULT,\
		.device = {\
			.class_id = #_name,\
			.class = &zt_sysfs_class,\
			CLASS_DEVICE_DEVT_HAVE(.devt = _dev,)\
			CLASS_DEVICE_RELEASE(.release = zt_sysfs_class_device_release,)\
		},\
		CLASS_DEVICE_DEVT_HAVENT(.devt = _dev,)\
}\



/* === Dynamically allocated Class device handling === */

/* simple declaration before the real definition (to avoid undefined errors) */
static struct class zt_sysfs_class;

/*
* @brief create a dynamically allocated zaptel class device
* 
* @param dev the underlying device if known
* @param devt the dev_t type if known or meaningfull
* @param data depends on type (for the moment either a zt_span or a zt_chan or NULL)
* @param type defines the type of this class device (one of ZT_CLASS_DEVICE_*)
* @param fmt  printf-like format string to give a name to this class_device
* @param ...  variables argument list that match fmt
*
* @return the pointer to newly created device or an error pointer
*/
static struct zt_sysfs_class_device *zt_sysfs_class_device_create(struct device *dev,
	dev_t devt, void *data, int type, const char *fmt, ...)
{
	va_list args;
	struct zt_sysfs_class_device *zcldev = NULL;
	int retval = -ENODEV;

	zcldev = kmalloc(sizeof(*zcldev), GFP_KERNEL);
    
	if (!zcldev) {
		retval = -ENOMEM;
		goto error;
	}
    
	memset(zcldev, 0, sizeof(*zcldev));
	/* be sure to set is_static field as 0 */
    
	zcldev->type = type;
    
	if (MAJOR(devt)) {
		CLASS_DEVICE_DEVT_HAVE(zcldev->device.devt = devt);
		CLASS_DEVICE_DEVT_HAVENT(zcldev->devt = devt);
	}
    
	zcldev->device.dev = dev;
	zcldev->device.class = &zt_sysfs_class;
    
	/* already provided in zt_sysfs_class but does not hurt */
	CLASS_DEVICE_RELEASE(zcldev->device.release = zt_sysfs_class_device_release);

	zcldev->device.class_data = data;
    
	va_start(args, fmt);
	vsnprintf(zcldev->device.class_id, BUS_ID_SIZE, fmt, args);
	va_end(args);
    
	retval = zt_class_device_register(zcldev);
	if (retval)
		goto error;

	return zcldev;
error:
	kfree(zcldev);
	return ERR_PTR(retval);
}


/* === Class device attributes handling === */

/*
* @brief get the name of function show (for an attribute)
*        it avoids name space clashes
*
* @param _name a base name that is unique for _type
* @param _type either chan or span
* 
*/  
#define ZT_ATTR_GET_SHOW(_name, _type)\
	zt_sysfs_##_type##_show_##_name
      
/*
* @brief define a show function (to use in a class_device attribute)
* for either a span or channel.
* 
* This only handles attributes to print a single expression result.
*
* @param _name name of this attribute
* @param _type either chan or span
* @param _typeup _type in upper-case
* @param _vartype the C type of the variable that can hold _expr
* @param _fmt the printf-like format string to print a _vartype
* @param _expr the expression that should be printed
*
* @note the resulting function has got the name ZT_ATTR_GET_SHOW(_name, _type)
*
* @note the _expr is executed under the chan->lock for a channel and under
*       span->lock for a span.
*
* @note class_device attributes are an overkill (for now -- kernel 2.6.19)
*       we have to define a separate function for each attribute...
*/
#define ZT_ATTR_DEFINE_SHOW_SIMPLE(_name, _type, _typeup, _vartype, _fmt, _expr)\
	static ssize_t	ZT_ATTR_GET_SHOW(_name, _type)(struct class_device *cldev,\
		char *buf)\
	{\
		struct zt_##_type *_type = (struct zt_##_type *)cldev->class_data;\
		unsigned long flags;\
		_vartype var;\
		\
		if (!_type) {\
			return 0;\
		}\
		\
		spin_lock_irqsave(&_type->lock, flags);\
		\
		if (ZT_INVALID_##_typeup##NO(_type->_type##no)) {\
			spin_unlock_irqrestore(&_type->lock, flags);\
			return 0;\
		}\
		\
		var = (_vartype)(_expr);\
		spin_unlock_irqrestore(&_type->lock, flags);\
		\
		return sprintf(buf, _fmt, var);\
	}
    

   
/*
* @brief define a read-only attribute that uses a show function
*        which name has been defined with ZT_ATTR_GET_SHOW
*/
#define ZT_ATTR_DEFINE_RO(_name, _type)\
	{\
		.attr = {.name = #_name, .mode = S_IRUGO, .owner = THIS_MODULE},\
		.show = ZT_ATTR_GET_SHOW(_name, _type),\
    }

/*
* @brief the NULL attribute (to terminate an array of class_device attributes)
*/    
#define ZT_ATTR_NULL \
	{.attr = {.name = NULL}}


/*
* @brief define a show function for a member of the zt_chan structure
*
* @param _member the member of the structure to print
* @param _vartype its C type
* @param _fmt a printf-like format that can print _vartype
*
* @see ZT_ATTR_DEFINE_SHOW_SIMPLE
*/
#define ZT_ATTR_CHAN_SHOW_MEMBER(_member, _vartype, _fmt)\
	ZT_ATTR_DEFINE_SHOW_SIMPLE(_member, chan, CHAN, _vartype, _fmt, chan->_member)
    
/*
* @brief define a read-only attribute that use a show function
*		which name has been defined with ZT_ATTR_GET_SHOW(*, chan)
*/
#define ZT_ATTR_CHAN_RO(_name)\
	ZT_ATTR_DEFINE_RO(_name, chan)

/*
* @brief define a show function for a member of the zt_span structure
*
* @param _member the member of the structure to print
* @param _vartype its C type
* @param _fmt a printf-like format that can print _vartype
*
* @see ZT_ATTR_DEFINE_SHOW_SIMPLE
*/
#define ZT_ATTR_SPAN_SHOW_MEMBER(_member, _vartype, _fmt)\
	ZT_ATTR_DEFINE_SHOW_SIMPLE(_member, span, SPAN, _vartype, _fmt, span->_member)
    
/*
* @brief define a read-only attribute that use a show function
*		which name has been defined with ZT_ATTR_GET_SHOW(*, span)
*/
#define ZT_ATTR_SPAN_RO(_name)\
	ZT_ATTR_DEFINE_RO(_name, span)



/* === Uevent/Hotplug helper macros === */

/*
* @brief add an uevent variable in an uevent environment buffer
*
* the variable will be called ZT_NAME (where NAME is _name)
*
* @param _name name (string)
* @param _fmt  printf-like format string for its value
* @param _val  value (any valid expression matching _fmt)
*
* @warning 
* - to be used by zt_sysfs_class_uevent only
*   (assumed variables envp, num_envp, i, length, buffer, buffer_size)
* - not to be added into a sub-block without '{' '}'
* -  may exit the function 
*/
#define ZT_UEVENT_VAR_ADD(_name, _fmt, _val)\
{\
	if (zt_sysfs_add_uevent_var(envp, num_envp, &i,\
		buffer, buffer_size, &length,\
		"ZT_" _name "=" _fmt, (_val)))\
		return -ENOMEM;\
}\

/*
* @brief terminate an uevent environment buffer
*
* @warning: 
* - to be used by zt_sysfs_class_uevent only
* (assumed variables envp, num_envp, i, length, buffer, buffer_size)
* - not to be added into a sub-block without '{' '}'
* - may exit the function
*/
#define ZT_UEVENT_VAR_END \
{\
		/* 
		* safety checking
		*/\
		if (i < num_envp)\
			envp[i] = NULL;\
		else\
			return -ENOMEM;\
}



/* ===  Various macros that are usefull === */

#define ZT_INVALID_SPANNO(_no)\
	((_no < 1) || (_no >= ZT_MAX_SPANS))\

#define ZT_INVALID_CHANNO(_no)\
	((_no < 1) || (_no >= ZT_MAX_CHANNELS))\


/* 
* @brief macro to get an error integer from a pointer 
* (NULL case beeing seen as -ENOMEM)
*/
#define ZT_PTR_ERR(pvoid)\
	((pvoid == NULL)? -ENOMEM : (IS_ERR(pvoid)? PTR_ERR(pvoid) : 0))
